/* Virtuino SE Json API example: Read - write digital IOs
 * Supported boards: Arduino UNO + ESP8266_01
 * Created by Ilias Lamprou
 * Updated Feb 27 2019
 * More examples: https://virtuino.com/index.php/virtuino-se/20-virtuino-se-code-examples
 */
// ------ VirtuinoSE V Memory Table
//  V0 memory -> digital output pin 13
//  V1 memory -> digital output pin 12
//  V2 memory -> digital input pin 6
//  V3 memory -> digital input pin 5
//  V4 memory -> reserved
//  V6 ... V15 memory -> reserved

#include <SoftwareSerial.h>
SoftwareSerial espSerial =  SoftwareSerial(2,3);      // arduino RX pin=2  arduino TX pin=3    connect the arduino RX pin to esp8266 module TX pin   -  connect the arduino TX pin to esp8266 module RX pin


#define outPin_1 13   
#define outPin_2 12 // Connect a led or a relay to this pin
#define inPin_1 6   // Connect a button to this pin
#define inPin_2 5   // Connect a button to this pin

//------------------- USER SETTINGS ---------------------------------
const char* ssid = "WIFI NETWORK";           // The name of your WiFi network (SSID)
const char* password = "PASSWORD";          // WIFI network PASSWORD
                                              // disable the line if you don't want to use a key                          
const char* serverIP = "192.168.1.150";       // The three first numbers have to be the same with the router IP

int port =8000;                               // Virtuino default Server port: 8000

const char* ssid_AP = "Virtuino network";     // Access point network SSID
const char* password_AP = "1234567890";       // Access point network PASSWORD

#define VIRTUINO_KEY  "1234"      // Virtuino KEY, only the requests that include this key will be acceptable. // disable the line if you don't want to use a key                          

#define V_memory_count 16         // the size of V memory. You can change it to a number <256)
//-------------------------------------------------------------------

float V_memory[V_memory_count];    // This array is synchronized with Virtuino SE V memory

boolean debug = true;              // set this variable to false on the finale code to decrease the request duration


//============================================================== connectToWiFiNetwork
// Remove this void if you want to connect your module as Access Point

void connectToWiFiNetwork(){
    Serial.println("Connecting to "+String(ssid));
    clearESP_buffer(1000);
    espSerial.println("AT+GMR");
    waitForResponse("OK",1000);
    
    espSerial.println("AT+CWMODE=1");  // configure as client
    waitForResponse("OK",1000);
    
    //--- connect
    espSerial.print("AT+CWJAP=\"");
    espSerial.print(ssid);
    espSerial.print("\",\"");
    espSerial.print(password);
    espSerial.println("\"");
    waitForResponse("OK",10000);

    espSerial.print("AT+CIPSTA=\"");
    espSerial.print(serverIP);
    espSerial.println("\"");   
    waitForResponse("OK",5000);

    espSerial.println("AT+CIPSTA?");
    waitForResponse("OK",3000); 
  
    espSerial.println("AT+CIFSR");           // get ip address
    waitForResponse("OK",1000);
    
    espSerial.println("AT+CIPMUX=1");         // configure for multiple connections   
    waitForResponse("OK",1000);
    
    espSerial.print("AT+CIPSERVER=1,");
    espSerial.println(port);
    waitForResponse("OK",1000);
   
}


//=================================================================
// Remove this void if you don't want to connect your module as Access Point

void initAccessPoint(){
      if (debug) Serial.println( "Init as AP...");
      clearESP_buffer(1000);
      espSerial.println("AT+GMR");
      waitForResponse("OK",1000);

      espSerial.println("AT+CWMODE=2");  // configure as client
      waitForResponse("OK",1000);
    
     espSerial.println("AT+CIFSR");
     waitForResponse("OK",1000);

     espSerial.print("AT+CWSAP=\"");
     espSerial.print(ssid_AP);
     espSerial.print("\",\"");
     espSerial.print(password_AP);
     espSerial.println("\",2,3");
     waitForResponse("OK",5000);
     
     espSerial.println("AT+CIPMUX=1");         // configure for multiple connections   
     waitForResponse("OK",2000);
    
     espSerial.print("AT+CIPSERVER=1,");
     espSerial.println(port);
     waitForResponse("OK",1000);           
}


//============================================================== setup
//==============================================================
void setup() {
  if (debug) {
    Serial.begin(9600);
    while (!Serial) continue;
  }
    espSerial.begin(9600);

  //initAccessPoint();               //enable this line to use the esp board as AP  
  connectToWiFiNetwork();      //enable this line to connect the module to your wifi network  
  
  pinMode(outPin_1, OUTPUT);            
  pinMode(outPin_2, OUTPUT);            
  pinMode(inPin_1, INPUT);
  pinMode(inPin_2, INPUT);
  
  }




//============================================================== loop
//==============================================================
void loop() {
  virtuinoRun();        //Necessary function to communicate with Virtuino. Client handler

  //--- How to control outputs from Virtuino SE
    if (V_memory[0]==1) digitalWrite(outPin_1,HIGH);      // On Virtuino SE panel add switch to control the pin outPin_1. On the switch settings select the memory V0
    else digitalWrite(outPin_1,LOW);

    if (V_memory[1]==1) digitalWrite(outPin_2,HIGH);      // On Virtuino SE panel add switch to control the pin outPin_2. On the switch settings select the memory V1
    else digitalWrite(outPin_2,LOW);
  
  
  //--- How to send to Virtuino SE the state of a pin every time it is changed 
  byte pinState = digitalRead(inPin_1);    // read the state of pin 6
  if (pinState!=V_memory[2]) {
      V_memory[2] = pinState;        // Store the pin state to V2. On Virtuino panel add a led to virtual memory V2  
  }

  pinState = digitalRead(inPin_2);    // read the state of pin 5
  if (pinState!=V_memory[3]) {
      V_memory[3] = pinState;        // Store the pin state to V3. On Virtuino panel add a led to virtual memory V3  
  }


}


//============================================================== onCommandReceived
//==============================================================
/* This function is called every time Virtuino SE app sends a request to server 
 * The 'variableIndex' is the V memory index of Virtuino app
 * The 'valueAsText' is the value that has sent from the app
 * If the 'valueAsText" is ? you have to return the value that you want to be displayed on Virtuino panel
 */
 String onCommandReceived(int variableIndex, String valueAsJson){     
   String valueAsText = getJsonValue(valueAsJson,"value");    // read the value from the received Json string 
   if (debug) Serial.println("New value received: V"+String(variableIndex)+"  value="+valueAsText);
 
  if (valueAsText=="?") return  String(V_memory[variableIndex]);   // return the value of the arduino V memory array
  else {
    float value = valueAsText.toFloat();        // convert the value to float. The valueAsText have to be numerical
    V_memory[variableIndex]=value;              // copy the received value to arduino V memory array
  }
  return "";
}
















//================== Virtuino Library (included in sketch) =====
//==============================================================
//==============================================================
/* Only the next functions are nedded to communicate with Virtuino 
 * Avoid to make changes to the code below  
 */


#define WRONG_KEY_MESSAGE  "Wrong Key"  
#define WELLCOME_MESSAGE  "Hello Virtuino"  

//============================================================== virtuinoRun
//==============================================================
void virtuinoRun(){
     if(espSerial.available()){
       delay(1);
       String espBuffer = espSerial.readStringUntil('\r');
          if (debug) Serial.print(espBuffer);
          int pos = espBuffer.indexOf("+IPD,");
          if (pos>=0){                                        // check for GET command from Virtuino app
              clearESP_buffer(100);                           // can be higher than 100
              int connectionId = espBuffer.charAt(pos+5)-48;  // get connection ID
              pos = espBuffer.indexOf(":");
              if (pos>0){                                   // We have a GET message
                   espBuffer=espBuffer.substring(pos+1);
                   String response= getResponseAsJson(espBuffer);                  
                    //---Send response to Virtuino SE
                    String cipSend = "AT+CIPSEND=";
                    cipSend += connectionId;
                    cipSend += ",";
                    cipSend += response.length();
                    cipSend += "\r\n";
                    espSerial.print(cipSend);
                    if (waitForResponse(">",2000)) espSerial.print(response);
              }
              delay(150);  // this delay is nedded
              espSerial.print("AT+CIPCLOSE=");
              espSerial.println(connectionId);
              
          }
          
     }

}

//============================================================== prepareResponse
//==============================================================
String getResponseAsJson(String data){
    //---check the KEY
   #ifdef VIRTUINO_KEY 
      String key = getJsonValue(data, "key");
      if (debug) Serial.println("Key="+key);
      if (!key.equals(VIRTUINO_KEY)) return "{\"status\":\"-1\",\"message\":\""+String(WRONG_KEY_MESSAGE)+"\"}";
   #endif

    //---check connection status
    String statusInfo = getJsonValue(data, "status");
    if (debug) Serial.println("statusInfo="+statusInfo);
    if (statusInfo.equals("0")) return "{\"status\":\"2\",\"message\":\""+String(WELLCOME_MESSAGE)+"\"}";
  
   String response="{\"status\":\"1\"";
   for (byte i=0;i<V_memory_count;i++){
      // easy way to read a Json object without installing extra json libraries
      String jsonTag="V"+String(i)+"_";
      int pos = data.indexOf(jsonTag);
      if (pos>0) {
        int startPos= data.indexOf("{",pos+1);
        if (startPos>pos) {
          int endPos= data.indexOf("}", startPos+2);
          if (endPos>startPos) {
            String variableAsJson = data.substring(startPos,endPos+1);
            String variableValue=onCommandReceived(i,variableAsJson);
           if (variableValue.length()>0) response+=",\""+jsonTag+"\":{\"value\":\""+urlencode(&variableValue)+"\"}";
          }
        }
      }
      
   } //for
   response+="}";
  return response;
}

//==============================================================
// easy way to read a Json text value without installing extra json libraries
  String getJsonValue(String jsonAsText, String jsonTag){
    int pos = jsonAsText.indexOf(jsonTag);
    if (pos>0) {
      int startPos=jsonAsText.indexOf(":\"",pos);
      int lastPos = jsonAsText.indexOf("\"",startPos+2);
      String value=jsonAsText.substring(startPos+2,lastPos);
      value=urldecode(&value);
      if (lastPos>startPos) return value;
    }
    return "";
 }
//======================================================================================== clearESP_buffer
//========================================================================================
void clearESP_buffer(int timeout){
    long t=millis();
    char c;
    while (t+timeout>millis()){
      if (espSerial.available()){
        c=espSerial.read();
        if (debug) Serial.print(c);
      }
    }
   }

//======================================================================================== waitForResponse
//========================================================================================
boolean waitForResponse(String target1,  int timeout){
    String data;
    char a;
    unsigned long start = millis();
    boolean rValue=false;
    while (millis() - start < timeout) {
        while(espSerial.available() > 0) {
            a = espSerial.read();
            if (debug) Serial.print(a);
            if(a == '\0') continue;
            data += a;
        }
        if (data.indexOf(target1) != -1) {
            rValue=true;
            break;
        } 
    }
    return rValue;
}

  //=================== ENCODE - DECODE Library (included in sketch) ==========
  //=================================================================
  
  //================================================================ urldecode
String urldecode(String* str){
   String encodedString="";
    char c;
    char code0;
    char code1;
    for (int i =0; i < str->length(); i++){
        c=str->charAt(i);
      if (c == '+'){
        encodedString+=' ';  
      }else if (c == '%') {
        i++;
        code0=str->charAt(i);
        i++;
        code1=str->charAt(i);
        c = (h2int(code0) << 4) | h2int(code1);
        encodedString+=c;
      } else{
         encodedString+=c;  
      }
      yield();
    }
    
   return encodedString;
}
//================================================================ urlencode
String urlencode(String* str){
   String encodedString="";
    char c;
    char code0;
    char code1;
    char code2;
    for (int i =0; i < str->length(); i++){
      c=str->charAt(i);
      if (c == ' '){
        encodedString+= '+';
      } else if (isalnum(c)){
        encodedString+=c;
      } else{
        code1=(c & 0xf)+'0';
        if ((c & 0xf) >9){
            code1=(c & 0xf) - 10 + 'A';
        }
        c=(c>>4)&0xf;
        code0=c+'0';
        if (c > 9){
            code0=c - 10 + 'A';
        }
        code2='\0';
        encodedString+='%';
        encodedString+=code0;
        encodedString+=code1;
        //encodedString+=code2;
      }
      yield();
    }
    return encodedString;
}
//================================================================ h2int
unsigned char h2int(char c){
    if (c >= '0' && c <='9'){
        return((unsigned char)c - '0');
    }
    if (c >= 'a' && c <='f'){
        return((unsigned char)c - 'a' + 10);
    }
    if (c >= 'A' && c <='F'){
        return((unsigned char)c - 'A' + 10);
    }
    return(0);
}



